package log

import (
	"bytes"
	"context"
	"encoding/base64"
	"flag"
	"fmt"
	"io/ioutil"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"text/template"
	"time"

	"github.com/hpcloud/tail"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

var log = _log.New("log")

// TODO add Timeout
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confInterval, "log.interval", confInterval, "[log] interval between polls")
}

type logDaemon struct {
	daemon.Daemon

	ttl             time.Duration
	defaultPatterns string
	patternPath     string
	namespace       string
	logs            []*logFile
	prefix          string

	wg    sync.WaitGroup
	stopC chan struct{}
}

type logService struct {
	common.Service
	Name  string
	Index int
	Type  string
}

type logFile struct {
	path     string
	service  *logService
	patterns []*logPattern
}

type logPattern struct {
	Type              string `yaml:"type"`
	Name              string `yaml:"name"`
	Pattern           string `yaml:"pattern"`
	regex             *regexp.Regexp
	responseTimeRatio float64
}
type logYaml map[string]map[string][]logPattern

var tailConfig = tail.Config{
	// Logger: tail.DiscardingLogger,
	Follow: true,
	// This prevents channel closure on file moves
	ReOpen: true,
	// This prevents tail from re-parsing previous lines
	Location: &tail.SeekInfo{Offset: 0, Whence: 2},
}

var serviceRegex = regexp.MustCompile(`^([^-]+)-(.+)-(\d+(?:\.\d+)?)$`)

func parsePatterns(path string, defaultPatterns string) (logYaml, error) {
	content := ""
	if path == "" {
		data, err := base64.StdEncoding.DecodeString(defaultPatterns)
		if err != nil {
			return nil, fmt.Errorf("Failed to base64-decode default patterns: %w", err)
		}
		content = string(data)
	} else {
		data, err := ioutil.ReadFile(path)
		if err != nil {
			return nil, fmt.Errorf("Failed to read file %s: %w", path, err)
		}
		content = string(data)
	}

	data := make(logYaml)
	if err := yaml.Unmarshal([]byte(content), &data); err != nil {
		return nil, fmt.Errorf("Failed to yaml-decode default patterns: %w", err)
	}

	return data, nil
}

func MakeDaemon(defaultPatterns, patternPath string, ttl time.Duration, namespace string) *logDaemon {
	return &logDaemon{
		ttl:             ttl,
		defaultPatterns: defaultPatterns,
		patternPath:     patternPath,
		namespace:       namespace,
		prefix:          path.Join("/var/log/oio/sds", namespace),
	}
}

func (d *logDaemon) ID() string {
	return "Logs"
}

func (d *logDaemon) Interval() time.Duration {
	return confInterval
}

func (d *logDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	return d.reload(services)
}

func (d *logDaemon) OnServicesChange(services common.ServicesMap) error {
	return d.reload(services)
}

func (d *logDaemon) reload(services common.ServicesMap) error {

	patterns, err := parsePatterns(d.patternPath, d.defaultPatterns)
	if err != nil {
		return fmt.Errorf("Failed to load patterns: %w", err)
	}
	log.Noticef("parsed log patterns for %d services", len(patterns))

	_logs := []*logFile{}

	for svcType, instances := range services {
		for _, svc := range instances {
			matches := serviceRegex.FindStringSubmatch(svc.ID)
			if matches == nil || len(matches) != 4 {
				log.Warnln("Could not expand service name for service", svc.ID)
				continue
			}

			serviceName := matches[2]
			serviceIndex, _ := strconv.Atoi(matches[3])

			service := logService{
				Service: *svc,
				Name:    serviceName,
				Index:   serviceIndex,
				Type:    svcType,
			}

			svcPatterns, ok := patterns[service.Type]
			if !ok {
				//TODO log debug when no pattern found
				continue
			}

			for pathTemplate, patterns := range svcPatterns {
				t := template.Must(template.New(service.ID).Parse(pathTemplate))
				var tpl bytes.Buffer
				if err := t.Execute(&tpl, service); err != nil {
					log.Warnf("error while templating logpath %s for service %s: %v", pathTemplate, service.ID, err)
					continue
				}
				lf := logFile{
					path:     path.Join(d.prefix, fmt.Sprintf("%s-%d", service.Type, service.Index), tpl.String()),
					service:  &service,
					patterns: []*logPattern{},
				}
				for _, pattern := range patterns {
					lp := pattern
					if lp.Type != "access" && lp.Type != "match" {
						return fmt.Errorf("Wrong pattern type (%s) for %s:%s", lp.Type, service.ID, lf.path)
					}
					re, err := regexp.Compile(lp.Pattern)
					if err != nil {
						return fmt.Errorf("Failed to compile pattern `%s` for service '%s': %w", lp.Pattern, service.ID, err)
					}
					lp.regex = re

					if lp.Type == "access" {
						// if the request_time field is a float (with a dot) the unit is in second
						// otherwise (the field is an int) the unit is in microsecond
						// and we want to export everything in seconds
						lp.responseTimeRatio = float64(1000000)
						for _, v := range strings.Split(lp.Pattern, " ") {
							if strings.Contains(v, "resp_time") {
								if strings.Contains(v, ".") {
									lp.responseTimeRatio = 1
								}
								break
							}
						}
					}

					if lp.Type == "match" {
						if lp.Name == "" {
							lp.Name = lp.Pattern
						}
					}

					lf.patterns = append(lf.patterns, &lp)
				}
				_logs = append(_logs, &lf)
			}
		}
	}

	log.Noticef("loaded %d logs for parsing", len(_logs))

	if d.stopC != nil {
		close(d.stopC)
		d.wg.Wait()
	}

	d.logs = _logs
	d.stopC = nil

	return nil
}

func (d *logDaemon) Collect(m *common.MetricData) error {
	if d.stopC != nil {
		return nil
	}

	d.wg = sync.WaitGroup{}
	d.stopC = make(chan struct{})

	for _, lf := range d.logs {
		d.wg.Add(1)
		go func(lf *logFile) {
			defer d.wg.Done()
			d.tailLogFile(lf, m)
		}(lf)
	}

	return nil
}

// TODO: If there is an error before we manage to watch the file, retry during the next Collect
func (d *logDaemon) tailLogFile(lf *logFile, m *common.MetricData) {
	if len(lf.patterns) == 0 {
		return
	}

	var t *tail.Tail
	for {
		var err error
		if t, err = tailFile(lf.path); err == nil {
			break
		}

		m.IncrError("log", lf.service.Type)
		log.Warnf("could not watch file '%s': %s. Retrying in %s", lf.path, err, d.ttl)

		select {
		case <-d.stopC:
			return
		case <-time.After(d.ttl):
		}
	}

	defer t.Cleanup()
	defer func() {
		if err := t.Stop(); err != nil {
			log.Warnf("error stopping tailing file '%s': %s", lf.path, err)
		}
	}()

	var line *tail.Line
	for {
		// Same trick as in the daemon package
		select {
		case <-d.stopC:
			return
		default:
		}

		select {
		case line = <-t.Lines:
		case <-d.stopC:
			return
		}

		if line.Err != nil {
			log.Warnf("error getting line: %s", line.Err)
			continue
		}

		for _, pattern := range lf.patterns {
			if err := d.handleLine(line.Text, pattern, lf.service, m); err != nil {
				m.IncrError("log", lf.service.Type)
				log.Warnf("error handling line from file '%s'(%s): %s", lf.path, pattern.Type, err)
			}
		}
	}
}

func tailFile(filePath string) (*tail.Tail, error) {
	if _, err := os.Stat(filePath); err != nil {
		return nil, err
	}
	return tail.TailFile(filePath, tailConfig)
}

func (d *logDaemon) handleLine(line string, pattern *logPattern, service *logService, m *common.MetricData) error {
	labels := [][]string{
		[]string{"service_addr", fmt.Sprintf("%s:%d", service.IP, service.Port)},
		[]string{"service", service.ID},
		[]string{"service_type", service.Type},
		[]string{"namespace", d.namespace},
	}
	if pattern.Type == "access" {
		return d.handleLineAccess(line, pattern, labels, m)
	}
	return d.handleLineMatch(line, pattern, labels, m)
}

func (d *logDaemon) handleLineMatch(line string, pattern *logPattern, labels [][]string, m *common.MetricData) error {
	if !pattern.regex.MatchString(line) {
		return nil
	}
	m.Acc(&common.Metric{
		Name:   "log_matches",
		Labels: append(labels, []string{"match", pattern.Name}),
		Value:  1,
		Type:   "counter",
		TTL:    d.ttl,
	})

	return nil
}

func (d *logDaemon) handleLineAccess(line string, pattern *logPattern, baseLabels [][]string, m *common.MetricData) error {

	s := pattern.regex.FindStringSubmatch(line)
	names := pattern.regex.SubexpNames()[1:]
	if len(s) != len(names)+1 {
		// TODO: add a debug line to see when a line is skipped
		return nil
	}

	if common.Index(names, "method") == -1 || common.Index(names, "code") == -1 {
		return nil
	}

	baseLabels = append(baseLabels, []string{"method", s[1+common.Index(names, "method")]})
	baseLabels = append(baseLabels, []string{"code", s[1+common.Index(names, "code")]})
	for idx, tag := range names {
		if strings.HasPrefix(tag, "tag_") {
			baseLabels = append(baseLabels, []string{strings.Split(tag, "tag_")[1], s[idx+1]})
		}
	}

	for idx, name := range names {
		if common.Index([]string{"bytes_in", "bytes_out", "resp_time", "code"}, name) == -1 {
			continue
		}

		value := s[idx+1]
		if value == "" || value == "-" {
			continue
		}

		vFloat, err := strconv.ParseFloat(value, 4)
		if err != nil {
			return fmt.Errorf("Could not convert '%s' to float: %w", name, err)
		}

		labels := make([][]string, len(baseLabels), len(baseLabels)+1)
		copy(labels, baseLabels)

		switch name {
		case "bytes_in":
			m.Acc(&common.Metric{
				Name:   "log_size_bytes",
				Labels: append(labels, []string{"direction", "in"}),
				Value:  vFloat,
				Type:   "stats",
				TTL:    d.ttl,
			})
		case "bytes_out":
			m.Acc(&common.Metric{
				Name:   "log_size_bytes",
				Labels: append(labels, []string{"direction", "out"}),
				Value:  vFloat,
				Type:   "stats",
				TTL:    d.ttl,
			})
		case "resp_time":
			m.Acc(&common.Metric{
				Name:   "log_duration_seconds",
				Labels: labels,
				Value:  float64(vFloat / pattern.responseTimeRatio),
				Type:   "stats",
				TTL:    d.ttl,
			})
		case "code":
			m.Acc(&common.Metric{
				Name:   "log_requests",
				Labels: labels,
				Value:  1,
				Type:   "counter",
				TTL:    d.ttl,
			})
		}
	}

	return nil
}

func (d *logDaemon) Close(ctx context.Context) error {
	if d.stopC == nil {
		return nil
	}

	close(d.stopC)

	done := make(chan struct{})
	go func() {
		d.wg.Wait()
		close(done)
	}()

	select {
	case <-done:
		return nil
	case <-ctx.Done():
		return ctx.Err()
	}
}
