package oiofs

import (
	"encoding/json"
	"flag"
	"fmt"
	"net/http"
	"strings"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("oiofs")

// TODO add Timeout
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confInterval, "oiofs.interval", confInterval, "[oiofs] interval between polls")
}

type oiofsDaemon struct {
	daemon.Daemon

	services  []*common.Service
	namespace string
}

func MakeDaemon(namespace string) *oiofsDaemon {
	return &oiofsDaemon{
		namespace: namespace,
	}
}

func (d *oiofsDaemon) ID() string {
	return "OIOFS"
}

func (d *oiofsDaemon) Interval() time.Duration {
	return confInterval
}

func (d *oiofsDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)

	return nil
}

func (d *oiofsDaemon) OnServicesChange(services common.ServicesMap) error {
	d.reload(services)

	return nil
}

func (d *oiofsDaemon) reload(services common.ServicesMap) {
	d.services = services["oiofs"]
}

func (d *oiofsDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup
	for _, svc := range d.services {
		wg.Add(1)
		go func(svc *common.Service) {
			defer wg.Done()

			if err := d.collect(svc, m); err != nil {
				m.IncrError("oiofs", "oiofs")
				log.Warnf("stats collection failed for service '%s': %s", svc.ID, err)
			}
		}(svc)
	}

	wg.Wait()

	return nil
}

func (d *oiofsDaemon) collect(svc *common.Service, m *common.MetricData) error {
	// TODO use common.HTTPGetJson function
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	res, err := http.Get("http://" + target + "/stats")
	if err != nil {
		return err
	}
	defer res.Body.Close()

	var data map[string]float64
	err = json.NewDecoder(res.Body).Decode(&data)
	if err != nil {
		return err
	}

	for k, v := range data {
		k = strings.ToLower(k)
		k = strings.Replace(k, "init_ctx", "initCtx", 1)
		parts := strings.Split(k, "_")

		metric := ""
		switch parts[len(parts)-1] {
		case "us":
			metric = "operation_time"
		case "count":
			metric = "operation_count"
		case "byte":
			metric = "bytes"
		case "microseconds":
			metric = "chunk_age"
		case "hit", "miss":
			metric = "cache_read"
		case "succeeded", "failed":
			// TODO Do something with this
			continue
		default:
			// TODO This shouldn't happen, add debug log
			continue
		}
		m.Set(&common.Metric{
			Name: "oiofs_" + metric,
			Labels: [][]string{
				[]string{"namespace", d.namespace},
				[]string{"service_type", "oiofs"},
				[]string{"service", svc.ID},
				[]string{"service_addr", target},
				[]string{"mountpoint", svc.Volume},
				[]string{"operation", strings.Trim(parts[1], " ")},
				[]string{"dimension", parts[2]},
				[]string{"module", parts[0]},
			},
			Value: v,
		})
	}

	return nil
}
