package openio

import (
	"errors"
	"flag"
	"fmt"
	"log"
	"os"
	"strconv"
	"strings"
	"sync"
	"syscall"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

type oioDaemon struct {
	daemon.Daemon

	proxyURL  string
	namespace string
	services  common.ServicesMap
}

type metaxInfoBody struct {
	Cache     map[string]int64
	Elections map[string]int64
}

type serviceInfoBody struct {
	Addr  string
	Score int
}

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "openio.timeout", confTimeout, "[openio] http timeout")
	fs.DurationVar(&confInterval, "openio.interval", confInterval, "[openio] interval between polls")
}

func MakeDaemon(namespace string) *oioDaemon {
	return &oioDaemon{
		namespace: namespace,
	}
}

func (d *oioDaemon) ID() string {
	return "OIO"
}

func (d *oioDaemon) Interval() time.Duration {
	return confInterval
}

func (d *oioDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	return d.reload(services)
}

func (d *oioDaemon) OnServicesChange(services common.ServicesMap) error {
	return d.reload(services)
}

func (d *oioDaemon) reload(services common.ServicesMap) error {
	d.services = services
	return nil
}

func (d *oioDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup

	svc, ok := d.services["oioproxy"]
	if !ok || len(svc) == 0 {
		return errors.New("no proxy found")
	}
	d.proxyURL = fmt.Sprintf("http://%s:%d/v3.0/", svc[0].IP, svc[0].Port)

	url := d.proxyURL + d.namespace + "/conscience/info?what=types"
	var types []string
	if err := common.HTTPGetJson(url, &types, confTimeout); err != nil {
		return fmt.Errorf("failed to fetch service types from conscience: %w", err)
	}

	for serviceType, services := range d.services {
		if common.Index(types, serviceType) == -1 {
			continue
		}

		wg.Add(1)
		go func(serviceType string, services []*common.Service) {
			defer wg.Done()

			d.collectScore(serviceType, services, m)
		}(serviceType, services)

		for _, svc := range services {
			wg.Add(2)
			go func(service *common.Service) {
				defer wg.Done()

				d.collectVolumeInfo(*service, m)
			}(svc)
			go func(serviceType string, service *common.Service) {
				defer wg.Done()

				switch serviceType {
				case "meta0", "meta1", "meta2":
					d.collectMeta(*service, m)
				case "rawx":
					d.collectRawx(*service, m)
				}
			}(serviceType, svc)
		}
	}

	wg.Wait()

	return nil
}

func (d *oioDaemon) collectMeta(svc common.Service, m *common.MetricData) {
	// TODO: move target as service address
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	labelsBase := [][]string{
		[]string{"namespace", svc.Namespace},
		[]string{"service", svc.ID},
		[]string{"service_type", svc.Type},
		[]string{"service_addr", target},
		[]string{"volume", svc.Volume},
	}

	counters, err := common.HTTPGetCsv(d.proxyURL+"forward/stats?id="+target, ' ', confTimeout)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN: metaX stats collection failed", err)
		return
	}
	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] != "counter" {
			continue
		}
		parts := strings.Split(cntr[1], ".")
		if len(parts) == 2 {
			parts = append(parts, "*")
		}
		if len(parts) != 3 {
			continue
		}

		var labels [][]string
		if parts[1] == "hits" || parts[1] == "time" {
			if parts[0] == "req" {
				labels = append(labelsBase, []string{"method", parts[2]})
			} else { // "rep"
				labels = append(labelsBase, []string{"code", parts[2]})
			}
		} else {
			labels = labelsBase
		}

		m.Set(&common.Metric{
			Name:   parts[0] + "_" + parts[1],
			Labels: labels,
			Value:  val,
		})
	}
	svcInfo := metaxInfoBody{}
	err = common.HTTPGetJson(d.proxyURL+"forward/info?id="+target, &svcInfo, confTimeout)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting metax info", svc.IP, svc.Port, err)
		return
	}
	for t, metrics := range map[string]map[string]int64{"base_cache_size/type": svcInfo.Cache, "base_elections/status": svcInfo.Elections} {
		vars := strings.Split(t, "/")
		if len(vars) != 2 {
			continue
		}
		for dim, val := range metrics {
			m.Set(&common.Metric{
				Name:   vars[0],
				Labels: append(labelsBase, []string{vars[1], dim}),
				Value:  float64(val),
			})
		}
	}
}

func (d *oioDaemon) collectRawx(svc common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	counters, err := common.HTTPGetCsv("http://"+target+"/stat", ' ', confTimeout)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN: rawx metric collection failed", err)
		return
	}
	labelsBase := [][]string{
		[]string{"namespace", svc.Namespace},
		[]string{"service", svc.ID},
		[]string{"service_type", svc.Type},
		[]string{"service_addr", target},
		[]string{"volume", svc.Volume},
	}
	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] != "counter" {
			continue
		}
		parts := strings.Split(cntr[1], ".")
		if len(parts) == 2 {
			parts = append(parts, "*")
		}
		if len(parts) != 3 {
			continue
		}

		var labels [][]string
		if parts[1] == "hits" || parts[1] == "time" {
			if parts[0] == "req" {
				labels = append(labelsBase, []string{"method", parts[2]})
			} else { // "rep"
				labels = append(labelsBase, []string{"code", parts[2]})
			}
		} else {
			labels = labelsBase
		}

		m.Set(&common.Metric{
			Name:   parts[0] + "_" + parts[1],
			Labels: labels,
			Value:  val,
		})
	}
}

func (d *oioDaemon) collectScore(svcType string, services []*common.Service, m *common.MetricData) {
	info := []serviceInfoBody{}
	err := common.HTTPGetJson(d.proxyURL+d.namespace+"/conscience/list?type="+svcType, &info, confTimeout)
	if err != nil {
		m.IncrError("openio", svcType)
		log.Println("WARN", "when collecting score info", svcType, err)
		return
	}
	for _, svc := range services {
		target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
		for _, remoteSvc := range info {
			if remoteSvc.Addr == target {
				m.Set(&common.Metric{
					Name: "score",
					Labels: [][]string{
						[]string{"namespace", d.namespace},
						[]string{"service", svc.ID},
						[]string{"service_type", svc.Type},
						[]string{"service_addr", target},
						[]string{"volume", svc.Volume},
					},
					Value: float64(remoteSvc.Score),
				})
				break
			}
		}
	}
}

func (d *oioDaemon) collectVolumeInfo(svc common.Service, m *common.MetricData) {
	if svc.Volume == "" {
		return
	}
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	f, err := os.OpenFile(svc.Volume, os.O_RDONLY, 0644)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting volume info", svc, err)
		return
	}
	defer f.Close()

	var stat syscall.Statfs_t
	err = syscall.Fstatfs(int(f.Fd()), &stat)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting volume info", svc, err)
		return
	}
	vMetric := map[string]uint64{
		"size_bytes/avail": stat.Bavail * uint64(stat.Bsize),
		"size_bytes/used":  (stat.Blocks - stat.Bfree) * uint64(stat.Bsize),
		"size_bytes/free":  stat.Bfree * uint64(stat.Bsize),
		"inodes/free":      stat.Ffree,
		"inodes/used":      stat.Files - stat.Ffree,
	}
	fsID := strconv.FormatUint(uint64(stat.Fsid.X__val[1])<<32|uint64(stat.Fsid.X__val[0]), 10)
	for dim, val := range vMetric {
		vars := strings.Split(dim, "/")
		if len(vars) != 2 {
			continue
		}

		m.Set(&common.Metric{
			Name: "volume_" + vars[0],
			Labels: [][]string{
				[]string{"namespace", d.namespace},
				[]string{"service", svc.ID},
				[]string{"service_type", svc.Type},
				[]string{"service_addr", target},
				[]string{"volume", svc.Volume},
				[]string{"fsid", fsID},
				[]string{"type", vars[1]},
			},
			Value: float64(val),
		})
	}
}
