package openio

import (
	"context"
	"fmt"
	"log"
	"net/http"
	"net/http/httptest"
	"strconv"
	"strings"
	"testing"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/testutils"
)

func makeFakeProxy(t *testing.T, rawxAddr, meta0Addr string) (*httptest.Server, string, int) {
	server := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var body string
		if r.URL.Path == "/v3.0/FAKENAMESPACE/conscience/info" && r.URL.RawQuery == "what=types" {
			body = `["meta0","rawx"]`
		} else if r.URL.Path == "/v3.0/FAKENAMESPACE/conscience/list" && r.URL.RawQuery == "type=rawx" {
			body = fmt.Sprintf(`[{"addr":"%s","score":54}]`, rawxAddr)
		} else if r.URL.Path == "/v3.0/FAKENAMESPACE/conscience/list" && r.URL.RawQuery == "type=meta0" {
			body = fmt.Sprintf(`[{"addr":"%s","score":67}]`, meta0Addr)
		} else if r.URL.Path == "/v3.0/forward/stats" && r.URL.RawQuery == fmt.Sprintf("id=%s", meta0Addr) {
			body = "counter req.hits.M0_ALL 123\ncounter req.time.M0_ALL 456"
		} else if r.URL.Path == "/v3.0/forward/info" && r.URL.RawQuery == fmt.Sprintf("id=%s", meta0Addr) {
			body = `{"elections":{"total":1,"none":2,"pending":3,"failed":4,"slave":5,"master":6},"cache":{"max":1,"soft_max":2,"hot":3,"cold":4,"used":5}}`
		}
		w.WriteHeader(200)
		_, _ = w.Write([]byte(body))
	}))

	parts := strings.Split(server.URL, ":")
	ip := parts[1][2:]
	port, err := strconv.ParseInt(parts[2], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return server, ip, int(port)
}

func makeFakeRawx(t *testing.T) (*httptest.Server, string, int) {
	server := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if r.URL.Path != "/stat" {
			return
		}
		w.WriteHeader(200)
		_, _ = w.Write([]byte("counter rep.hits.2xx 123\n"))
		_, _ = w.Write([]byte("counter rep.hits.4xx 456\n"))
		_, _ = w.Write([]byte("config service_id x.x.x.x:6200\n"))
	}))

	parts := strings.Split(server.URL, ":")
	ip := parts[1][2:]
	port, err := strconv.ParseInt(parts[2], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return server, ip, int(port)
}

func TestOpenioModule(t *testing.T) {
	t.Parallel()

	m := common.MakeMetricData()
	d := MakeDaemon("FAKENAMESPACE")

	rawxServer, rawxIP, rawxPort := makeFakeRawx(t)
	defer rawxServer.Close()
	rawxAddr := fmt.Sprintf("%s:%v", rawxIP, rawxPort)

	meta0Addr := "0.0.0.0:6001"

	proxyServer, proxyIP, proxyPort := makeFakeProxy(t, rawxAddr, meta0Addr)
	defer proxyServer.Close()

	services := common.ServicesMap{
		"oioproxy": []*common.Service{
			{
				IP:   proxyIP,
				Port: proxyPort,
			},
		},
		"rawx": []*common.Service{
			{
				ID:   "fake-rawx-0",
				IP:   rawxIP,
				Port: rawxPort,
				Type: "rawx",
			},
		},
		"meta0": []*common.Service{
			{
				ID:   "fake-meta0-0",
				IP:   "0.0.0.0",
				Port: 6001,
				Type: "meta0",
			},
		},
	}

	if err := d.Init(services, nil); err != nil {
		t.Fatalf("Init error: %s", err)
	}

	if err := d.Collect(m); err != nil {
		t.Fatalf("Collect error: %s", err)
	}

	if err := d.Close(context.TODO()); err != nil {
		t.Fatalf("Close error: %s", err)
	}

	testutils.NewMetricDataTest().
		ForAllMetrics(
			testutils.NewMetricAssertion().HasLabels("service", "service_type", "service_addr")).
		ForMetricsWithLabels(
			testutils.NewMetricAssertion().
				LabelIs("service_type", "rawx").
				LabelIs("service_addr", rawxAddr),
			"service", "fake-rawx-0").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(54),
			"score",
			"service", "fake-rawx-0").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(123).
				LabelIs("service", "fake-rawx-0"),
			"rep_hits",
			"code", "2xx").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(456).
				LabelIs("service", "fake-rawx-0"),
			"rep_hits",
			"code", "4xx").
		ForMetricsWithLabels(
			testutils.NewMetricAssertion().
				LabelIs("service_type", "meta0").
				LabelIs("service_addr", meta0Addr),
			"service", "fake-meta0-0").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(67),
			"score",
			"service", "fake-meta0-0").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(123).
				LabelIs("service", "fake-meta0-0"),
			"req_hits",
			"method", "M0_ALL").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(456).
				LabelIs("service", "fake-meta0-0"),
			"req_time",
			"method", "M0_ALL").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(5).
				LabelIs("service", "fake-meta0-0"),
			"base_elections",
			"status", "slave").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(6).
				LabelIs("service", "fake-meta0-0"),
			"base_elections",
			"status", "master").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(3).
				LabelIs("service", "fake-meta0-0"),
			"base_cache_size",
			"type", "hot").
		ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(4).
				LabelIs("service", "fake-meta0-0"),
			"base_cache_size",
			"type", "cold").
		NoExtraMetrics().
		Check(t, m)
}
