package testutils

import (
	"fmt"
	"testing"

	"github.com/open-io/oio-exporter/common"
)

type MetricDataTest struct {
	metricDataAsserts []*metricDataAssertion
	metrics           map[string]struct{}
	noExtraMetrics    bool
}

func NewMetricDataTest() *MetricDataTest {
	return &MetricDataTest{
		metricDataAsserts: make([]*metricDataAssertion, 0),
		metrics:           make(map[string]struct{}),
	}
}

type metricDataAssertion struct {
	matchers     []metricMatcherI
	metricAssert *metricAssertion
}

func newMetricDataAssertion(metricAssert *metricAssertion) *metricDataAssertion {
	return &metricDataAssertion{
		matchers:     make([]metricMatcherI, 0),
		metricAssert: metricAssert,
	}
}

func (mda *metricDataAssertion) Match(metric *common.Metric) bool {
	for _, matcher := range mda.matchers {
		if !matcher.Match(metric) {
			return false
		}
	}
	return true
}

func (mda *metricDataAssertion) MatchingMetrics(measures common.Measures) []*common.Metric {
	matchingMetrics := make([]*common.Metric, 0)
	for _, metric := range measures {
		if mda.Match(metric) {
			matchingMetrics = append(matchingMetrics, metric)
		}
	}
	return matchingMetrics
}

func (mda *metricDataAssertion) Name(metricName string) *metricDataAssertion {
	mda.matchers = append(mda.matchers, &nameMetricMatcher{metricName})
	return mda
}

func (mda *metricDataAssertion) Label(labelName, labelValue string) *metricDataAssertion {
	mda.matchers = append(mda.matchers, &labelMetricMatcher{labelName, labelValue})
	return mda
}

type metricMatcherI interface {
	Match(*common.Metric) bool
}

type nameMetricMatcher struct {
	name string
}

func (mm *nameMetricMatcher) Match(metric *common.Metric) bool {
	return metric.Name == mm.name
}

func (mm *nameMetricMatcher) String() string {
	return fmt.Sprintf("[name: %s]", mm.name)
}

type labelMetricMatcher struct {
	name, value string
}

func (mm *labelMetricMatcher) Match(metric *common.Metric) bool {
	for _, label := range metric.Labels {
		if label[0] == mm.name && label[1] == mm.value {
			return true
		}
	}

	return false
}

func (mm *labelMetricMatcher) String() string {
	return fmt.Sprintf("[label: %s/%s]", mm.name, mm.value)
}

func (mdt *MetricDataTest) ForAllMetrics(metricAssert *metricAssertion) *MetricDataTest {
	mdt.metricDataAsserts = append(mdt.metricDataAsserts, newMetricDataAssertion(metricAssert))
	return mdt
}

func (mdt *MetricDataTest) ForMetric(metricAssert *metricAssertion, metricName string) *MetricDataTest {
	return mdt.ForMetricWithLabels(metricAssert, metricName)
}

func (mdt *MetricDataTest) ForMetricWithLabels(metricAssert *metricAssertion, metricName string, labels ...string) *MetricDataTest {
	mda := newMetricDataAssertion(metricAssert).Name(metricName)
	for i := 0; i < len(labels); i += 2 {
		mda.Label(labels[i], labels[i+1])
	}
	mdt.metrics[metricName] = struct{}{}
	mdt.metricDataAsserts = append(mdt.metricDataAsserts, mda)
	return mdt
}

func (mdt *MetricDataTest) ForMetricsWithLabels(metricAssert *metricAssertion, labels ...string) *MetricDataTest {
	mda := newMetricDataAssertion(metricAssert)
	for i := 0; i < len(labels); i += 2 {
		mda.Label(labels[i], labels[i+1])
	}
	mdt.metricDataAsserts = append(mdt.metricDataAsserts, mda)
	return mdt
}

// Convenience function to declare required metrics without any assertions
func (mdt *MetricDataTest) HasMetrics(metricNames ...string) *MetricDataTest {
	for _, metricName := range metricNames {
		mdt.ForMetricWithLabels(&metricAssertion{}, metricName)
	}
	return mdt
}

// Only metrics with an assertion defined are allowed
func (mdt *MetricDataTest) NoExtraMetrics() *MetricDataTest {
	mdt.noExtraMetrics = true
	return mdt
}

type metricAssertion struct {
	valueAssert   *valueAssertion
	labels        map[string]struct{}
	labelAsserts  []*labelAssertion
	noExtraLabels bool
}

func NewMetricAssertion() *metricAssertion {
	return &metricAssertion{
		labels:       make(map[string]struct{}),
		labelAsserts: make([]*labelAssertion, 0),
	}
}

type valueAssertion struct {
	operator valueOperator
	value    float64
}

type valueOperator int

const (
	valueLT valueOperator = iota
	valueLTE
	valueE
	valueGTE
	valueGT
)

func (ma *metricAssertion) setValueAssertion(operator valueOperator, value float64) {
	ma.valueAssert = &valueAssertion{
		operator: operator,
		value:    value,
	}
}

func (ma *metricAssertion) LowerThan(value float64) *metricAssertion {
	ma.setValueAssertion(valueLT, value)
	return ma
}
func (ma *metricAssertion) LowerThanOrEquals(value float64) *metricAssertion {
	ma.setValueAssertion(valueLTE, value)
	return ma
}
func (ma *metricAssertion) Equals(value float64) *metricAssertion {
	ma.setValueAssertion(valueE, value)
	return ma
}
func (ma *metricAssertion) GreaterThanOrEquals(value float64) *metricAssertion {
	ma.setValueAssertion(valueGTE, value)
	return ma
}
func (ma *metricAssertion) GreaterThan(value float64) *metricAssertion {
	ma.setValueAssertion(valueGT, value)
	return ma
}

type labelAssertion struct {
	operator    labelOperator
	name, value string
}

type labelOperator int

const (
	hasLabel labelOperator = iota
	labelIs
)

func (ma *metricAssertion) HasLabel(label string) *metricAssertion {
	ma.labels[label] = struct{}{}
	ma.labelAsserts = append(ma.labelAsserts, &labelAssertion{
		operator: hasLabel,
		name:     label,
	})
	return ma
}

func (ma *metricAssertion) HasLabels(labels ...string) *metricAssertion {
	for _, label := range labels {
		ma.HasLabel(label)
	}
	return ma
}

func (ma *metricAssertion) LabelIs(label, value string) *metricAssertion {
	ma.labels[label] = struct{}{}
	ma.labelAsserts = append(ma.labelAsserts, &labelAssertion{
		operator: labelIs,
		name:     label,
		value:    value,
	})
	return ma
}

// Only labels with an assertion defined are allowed
func (ma *metricAssertion) NoExtraLabels() *metricAssertion {
	ma.noExtraLabels = true
	return ma
}

func (mdt MetricDataTest) Check(t *testing.T, m *common.MetricData) {
	if mdt.noExtraMetrics {
		mdt.ensureNoExtraMetrics(t, m)
	}

	for _, mda := range mdt.metricDataAsserts {
		mda.check(t, m)
	}
}

func (mdt MetricDataTest) ensureNoExtraMetrics(t *testing.T, m *common.MetricData) {
	for _, metric := range m.Measures {
		metricName := metric.Name

		if _, ok := mdt.metrics[metricName]; !ok {
			t.Fatalf("Unexpected metric '%s'", metricName)
		}
	}
}

func (mda *metricDataAssertion) check(t *testing.T, m *common.MetricData) {
	matchingMetrics := mda.MatchingMetrics(m.Measures)
	if len(matchingMetrics) == 0 {
		t.Fatalf("No metric matching: %s", mda.matchers)
	}
	ctx := fmt.Sprint(mda.matchers)
	for _, metric := range matchingMetrics {
		mda.metricAssert.check(t, ctx, metric)
	}
}

func (ma *metricAssertion) check(t *testing.T, ctx string, metric *common.Metric) {
	if ma.valueAssert != nil {
		ma.valueAssert.check(t, ctx, metric.Value)
	}

	if ma.noExtraLabels {
		ma.ensureNoExtraLabels(t, ctx, metric.Labels)
	}

	for _, labelAssert := range ma.labelAsserts {
		labelAssert.check(t, ctx, metric)
	}
}

func (va *valueAssertion) check(t *testing.T, ctx string, value float64) {
	switch va.operator {
	case valueLT:
		if value >= va.value {
			t.Fatalf("Expected metric %s value to be lower than %v, got: %v", ctx, va.value, value)
		}
	case valueLTE:
		if value > va.value {
			t.Fatalf("Expected metric %s value to be lower than or equal to %v, got: %v", ctx, va.value, value)
		}
	case valueE:
		if value != va.value {
			t.Fatalf("Expected metric %s value to be equal to %v, got: %v", ctx, va.value, value)
		}
	case valueGTE:
		if value < va.value {
			t.Fatalf("Expected metric %s value to be greater than or equal to %v, got: %v", ctx, va.value, value)
		}
	case valueGT:
		if value <= va.value {
			t.Fatalf("Expected metric %s value to be greater than %v, got: %v", ctx, va.value, value)
		}
	default:
		t.Fatalf("Undefined valueAssertion operator: %d", va.operator)
	}
}

func (ma *metricAssertion) ensureNoExtraLabels(t *testing.T, ctx string, labels [][]string) {
	for _, label := range labels {
		labelName := label[0]

		if _, ok := ma.labels[labelName]; !ok {
			t.Fatalf("Unexpected label '%s' for metric %s", labelName, ctx)
		}
	}
}

func (la *labelAssertion) check(t *testing.T, ctx string, metric *common.Metric) {
	labelName := la.name

	labelValue, ok := metric.GetLabel(labelName)
	if !ok {
		t.Fatalf("Missing label '%s' for metric %s", labelName, ctx)
	}

	if la.operator == labelIs && la.value != labelValue {
		t.Fatalf("Expected label '%s' of metric %s to be %s, got: %s", labelName, ctx, la.value, labelValue)
	}
}
