package zookeeper

import (
	"flag"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/common/tcp"
)

var log = _log.New("zookeeper")

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "zookeeper.timeout", confTimeout, "[zookeeper] operation timeout")
	fs.DurationVar(&confInterval, "zookeeper.interval", confInterval, "[zookeeper] interval between polls")
}

func MakeDaemon(namespace string) *tcp.TcpDaemon {
	return &tcp.TcpDaemon{
		Timeout:        confTimeout,
		ConfInterval:   confInterval,
		Namespace:      namespace,
		ServiceType:    "zookeeper",
		Keepalive:      false,
		CollectService: collect,
	}
}

func collect(d *tcp.TcpDaemon, svc *common.Service, m *common.MetricData, target string) {
	data, err := d.Command(svc, "mntr\n", []string{""})
	if err != nil {
		m.IncrError("zookeeper", "zookeeper")
		log.Warnln("error while collecting zookeeper metrics", svc, err)
		return
	}

	for _, line := range data {
		kv := strings.SplitN(line, "\t", 2)
		if len(kv) != 2 {
			// TODO log
			continue
		}
		stat := strings.Replace(kv[0], "zk_", "", 1)
		valStr := kv[1]
		val, errFloat := strconv.ParseFloat(valStr, 64)

		labels := [][]string{
			[]string{"namespace", d.Namespace},
			[]string{"service", svc.ID},
			[]string{"service_type", svc.Type},
			[]string{"service_addr", target},
		}

		switch stat {

		case "avg_latency":
			stat = "latency_seconds_avg"
			/* convert to seconds */
			if errFloat == nil {
				val /= 1000
			}

		case "min_latency":
			stat = "latency_seconds_min"
			/* convert to seconds */
			if errFloat == nil {
				val /= 1000
			}

		case "max_latency":
			stat = "latency_seconds_max"
			/* convert to seconds */
			if errFloat == nil {
				val /= 1000
			}

		case "packets_received":
			stat = "packets"
			labels = append(labels, []string{"direction", "in"})

		case "packets_sent":
			stat = "packets"
			labels = append(labels, []string{"direction", "out"})

		case "num_alive_connections":
			stat = "alive_connections"

		case "server_state":
			labels = append(labels, []string{"role", valStr})
			val = float64(time.Now().Unix())
			errFloat = nil

		case "approximate_data_size":
			stat = "approximate_data_size_bytes"

		case "open_file_descriptor_count":
			stat = "open_files"

		case "max_file_descriptor_count":
			stat = "open_files_max"

		case "znode_count":
			stat = "items"
			labels = append(labels, []string{"type", "znode"})

		case "watch_count":
			stat = "items"
			labels = append(labels, []string{"type", "watch"})

		case "ephemerals_count":
			stat = "items"
			labels = append(labels, []string{"type", "ephemerals"})

		case "last_proposal_size":
			stat = "proposal_size"

		case "min_proposal_size":
			stat = "proposal_size_min"

		case "max_proposal_size":
			stat = "proposal_size_max"

		/* ignore useless metrics */
		case "version":
			continue

		/* by default, keep not listed metrics without changing its name and considering it's a float value */
		default:
		}

		if errFloat != nil {
			// TODO log
			continue
		}

		m.Set(&common.Metric{
			Name:   "zookeeper_" + stat,
			Labels: labels,
			Value:  val,
		})
	}
}
